@@ -2,29 +2,24 @@
 Proxies for libgeos, GEOS-specific exceptions, and utilities
 """
 
+import atexit
+from ctypes import (
+    CDLL, cdll, pointer, string_at, DEFAULT_MODE, c_void_p, c_size_t, c_char_p)
+from ctypes.util import find_library
+import glob
+import logging
 import os
 import re
 import sys
-import atexit
-import logging
 import threading
-from ctypes import CDLL, cdll, pointer, string_at, cast, POINTER, DEFAULT_MODE
-from ctypes import c_void_p, c_size_t, c_char_p, c_int, c_float
-from ctypes.util import find_library
 
 from .ctypes_declarations import prototype, EXCEPTION_HANDLER_FUNCTYPE
+from .errors import WKBReadingError, WKTReadingError, TopologicalError, PredicateError
 from . import ftools
 
 
 # Add message handler to this module's logger
 LOG = logging.getLogger(__name__)
-ch = logging.StreamHandler()
-LOG.addHandler(ch)
-
-if 'all' in sys.warnoptions:
-    # show GEOS messages in console with: python -W all
-    LOG.setLevel(logging.DEBUG)
-
 
 # Find and load the GEOS and C libraries
 # If this ever gets any longer, we'll break it into separate modules
@@ -37,7 +32,7 @@ def load_dll(libname, fallbacks=None, mode=DEFAULT_MODE):
             LOG.debug("Trying `CDLL(%s)`", lib)
             dll = CDLL(lib, mode=mode)
         except OSError:
-            LOG.warn("Failed `CDLL(%s)`", lib)
+            LOG.debug("Failed `CDLL(%s)`", lib)
             pass
 
     if not dll and fallbacks is not None:
@@ -47,7 +42,7 @@ def load_dll(libname, fallbacks=None, mode=DEFAULT_MODE):
                 dll = CDLL(name, mode=mode)
             except OSError:
                 # move on to the next fallback
-                LOG.warn("Failed `CDLL(%s)`", name)
+                LOG.debug("Failed `CDLL(%s)`", name)
                 pass
 
     if dll:
@@ -63,7 +58,21 @@ def load_dll(libname, fallbacks=None, mode=DEFAULT_MODE):
 _lgeos = None
 
 if sys.platform.startswith('linux'):
-    _lgeos = load_dll('geos_c', fallbacks=['libgeos_c.so.1', 'libgeos_c.so'])
+    # Test to see if we have a wheel repaired by 'auditwheel' containing its
+    # own libgeos_c
+    geos_whl_so = glob.glob(os.path.abspath(os.path.join(os.path.dirname(
+        __file__), '.libs/libgeos_c-*.so.*')))
+    if len(geos_whl_so) == 1:
+        _lgeos = CDLL(geos_whl_so[0])
+        LOG.debug("Found GEOS DLL: %r, using it.", _lgeos)
+    else:
+        alt_paths = [
+            'libgeos_c.so.1',
+            'libgeos_c.so',
+            # anaconda
+            os.path.join(sys.prefix, "lib", "libgeos_c.so"),
+        ]
+        _lgeos = load_dll('geos_c', fallbacks=alt_paths)
     free = load_dll('c').free
     free.argtypes = [c_void_p]
     free.restype = None
@@ -80,14 +89,20 @@ elif sys.platform == 'darwin':
         if hasattr(sys, 'frozen'):
             try:
                 # .app file from py2app
-                alt_paths = [os.path.join(os.environ['RESOURCEPATH'],
-                            '..', 'Frameworks', 'libgeos_c.dylib')]
+                alt_paths = [os.path.join(
+                    os.environ['RESOURCEPATH'], '..', 'Frameworks',
+                    'libgeos_c.dylib')]
             except KeyError:
                 # binary from pyinstaller
                 alt_paths = [
                     os.path.join(sys.executable, 'libgeos_c.dylib')]
+                if hasattr(sys, '_MEIPASS'):
+                    alt_paths.append(
+                        os.path.join(sys._MEIPASS, 'libgeos_c.1.dylib'))
         else:
             alt_paths = [
+                # anaconda
+                os.path.join(sys.prefix, "lib", "libgeos_c.dylib"),
                 # The Framework build from Kyng Chaos
                 "/Library/Frameworks/GEOS.framework/Versions/Current/GEOS",
                 # macports
@@ -105,13 +120,15 @@ elif sys.platform == 'win32':
             os.path.join(os.path.dirname(__file__), 'DLLs'))
         if hasattr(sys, "frozen"):
             wininst_dlls = os.path.normpath(
-                os.path.abspath(sys.executable+'../../DLLS'))
+                os.path.abspath(sys.executable + '../../DLLS'))
         else:
             wininst_dlls = os.path.abspath(os.__file__ + "../../../DLLs")
         original_path = os.environ['PATH']
         os.environ['PATH'] = "%s;%s;%s" % \
             (egg_dlls, wininst_dlls, original_path)
-        _lgeos = CDLL("geos.dll")
+        _lgeos = load_dll("geos_c.dll", fallbacks=[
+            os.path.join(sys.prefix, "Library", "lib", "geos_c.dll"),
+        ])
     except (ImportError, WindowsError, OSError):
         raise
 
@@ -135,11 +152,9 @@ else:  # other *nix systems
 
 
 def _geos_version():
-    # extern const char GEOS_DLL *GEOSversion();
     GEOSversion = _lgeos.GEOSversion
     GEOSversion.restype = c_char_p
     GEOSversion.argtypes = []
-    #define GEOS_CAPI_VERSION "@VERSION@-CAPI-@CAPI_VERSION@"
     geos_version_string = GEOSversion()
     if sys.version_info[0] >= 3:
         geos_version_string = geos_version_string.decode('ascii')
@@ -187,51 +202,26 @@ if geos_version >= (3, 1, 0):
         [EXCEPTION_HANDLER_FUNCTYPE, EXCEPTION_HANDLER_FUNCTYPE]
     _lgeos.finishGEOS_r.argtypes = [c_void_p]
 
-# Exceptions
 
+def make_logging_callback(func):
+    """Error or notice handler callback producr
 
-class ReadingError(Exception):
-    pass
-
-
-class DimensionError(Exception):
-    pass
-
-
-class TopologicalError(Exception):
-    pass
-
-
-class PredicateError(Exception):
-    pass
-
-# While this function can take any number of positional arguments when
-# called from Python and GEOS expects its error handler to accept any
-# number of arguments (like printf), I'm unable to get ctypes to make
-# a callback object from this function that will accept any number of
-# arguments.
-#
-# At the moment, functions in the GEOS C API only pass 0 or 1 arguments
-# to the error handler. We can deal with this, but when if that changes,
-# Shapely may break.
-
-def handler(level):
-    def callback(fmt, *args):
+    Wraps a logger method, func, as a GEOS callback.
+    """
+    def callback(fmt, *fmt_args):
         fmt = fmt.decode('ascii')
         conversions = re.findall(r'%.', fmt)
-        log_vals = []
-        for spec, arg in zip(conversions, args):
-            if spec == '%s' and arg is not None:
-                log_vals.append(string_at(arg).decode('ascii'))
-            else:
-                LOG.error("An error occurred, but the format string "
-                          "'%s' could not be converted.", fmt)
-                return
-        getattr(LOG, level)(fmt, *log_vals)
+        args = [
+            string_at(arg).decode('ascii')
+            for spec, arg in zip(conversions, fmt_args)
+            if spec == '%s' and arg is not None]
+
+        func(fmt, *args)
+
     return callback
 
-error_handler = handler('error')
-notice_handler = handler('warning')
+error_handler = make_logging_callback(LOG.error)
+notice_handler = make_logging_callback(LOG.info)
 
 error_h = EXCEPTION_HANDLER_FUNCTYPE(error_handler)
 notice_h = EXCEPTION_HANDLER_FUNCTYPE(notice_handler)
@@ -260,8 +250,9 @@ class WKTReader(object):
             text = text.encode('ascii')
         geom = self._lgeos.GEOSWKTReader_read(self._reader, c_char_p(text))
         if not geom:
-            raise ReadingError("Could not create geometry because of errors "
-                               "while reading input.")
+            raise WKTReadingError(
+                "Could not create geometry because of errors "
+                "while reading input.")
         # avoid circular import dependency
         from shapely.geometry.base import geom_factory
         return geom_factory(geom)
@@ -400,8 +391,9 @@ class WKBReader(object):
         geom = self._lgeos.GEOSWKBReader_read(
             self._reader, c_char_p(data), c_size_t(len(data)))
         if not geom:
-            raise ReadingError("Could not create geometry because of errors "
-                               "while reading input.")
+            raise WKBReadingError(
+                "Could not create geometry because of errors "
+                "while reading input.")
         # avoid circular import dependency
         from shapely import geometry
         return geometry.base.geom_factory(geom)
@@ -413,8 +405,9 @@ class WKBReader(object):
         geom = self._lgeos.GEOSWKBReader_readHEX(
             self._reader, c_char_p(data), c_size_t(len(data)))
         if not geom:
-            raise ReadingError("Could not create geometry because of errors "
-                               "while reading input.")
+            raise WKBReadingError(
+                "Could not create geometry because of errors "
+                "while reading input.")
         # avoid circular import dependency
         from shapely import geometry
         return geometry.base.geom_factory(geom)
@@ -535,13 +528,20 @@ def errcheck_just_free(result, func, argtuple):
     else:
         return retval
 
+
 def errcheck_null_exception(result, func, argtuple):
-    """Wraps errcheck_just_free, raising a TopologicalError if result is NULL"""
+    """Wraps errcheck_just_free
+
+    Raises TopologicalError if result is NULL.
+    """
     if not result:
-        raise TopologicalError("The operation '{0}' could not be performed."
-            "Likely cause is invalidity of the geometry.".format(func.__name__))
+        raise TopologicalError(
+            "The operation '{0}' could not be performed."
+            "Likely cause is invalidity of the geometry.".format(
+                func.__name__))
     return errcheck_just_free(result, func, argtuple)
 
+
 def errcheck_predicate(result, func, argtuple):
     """Result is 2 on exception, 1 on True, 0 on False"""
     if result == 2:
@@ -758,12 +758,16 @@ class LGEOS320(LGEOS311):
         super(LGEOS320, self).__init__(dll)
 
         if geos_version >= (3, 2, 0):
-            def parallel_offset(geom, distance, resolution=16, join_style=1, mitre_limit=5.0, side='right'):
+
+            def parallel_offset(geom, distance, resolution=16, join_style=1,
+                                mitre_limit=5.0, side='right'):
                 side = side == 'left'
                 if distance < 0:
                     distance = abs(distance)
                     side = not side
-                return self.GEOSSingleSidedBuffer(geom, distance, resolution, join_style, mitre_limit, side)
+                return self.GEOSSingleSidedBuffer(
+                    geom, distance, resolution, join_style, mitre_limit, side)
+
             self.methods['parallel_offset'] = parallel_offset
 
         self.methods['project'] = self.GEOSProject
@@ -772,6 +776,7 @@ class LGEOS320(LGEOS311):
         self.methods['interpolate_normalized'] = \
             self.GEOSInterpolateNormalized
         self.methods['buffer_with_style'] = self.GEOSBufferWithStyle
+        self.methods['hausdorff_distance'] = self.GEOSHausdorffDistance
 
 
 class LGEOS330(LGEOS320):
@@ -794,16 +799,19 @@ class LGEOS330(LGEOS320):
         for pred in (self.GEOSisClosed,):
             pred.func.errcheck = errcheck_predicate
 
-        def parallel_offset(geom, distance, resolution=16, join_style=1, mitre_limit=5.0, side='right'):
+        def parallel_offset(geom, distance, resolution=16, join_style=1,
+                            mitre_limit=5.0, side='right'):
             if side == 'right':
                 distance *= -1
-            return self.GEOSOffsetCurve(geom, distance, resolution, join_style, mitre_limit)
-        self.methods['parallel_offset'] = parallel_offset
+            return self.GEOSOffsetCurve(
+                geom, distance, resolution, join_style, mitre_limit)
 
+        self.methods['parallel_offset'] = parallel_offset
         self.methods['unary_union'] = self.GEOSUnaryUnion
         self.methods['is_closed'] = self.GEOSisClosed
         self.methods['cascaded_union'] = self.methods['unary_union']
         self.methods['snap'] = self.GEOSSnap
+        self.methods['shared_paths'] = self.GEOSSharedPaths
 
 
 class LGEOS340(LGEOS330):
